/*
 * cgi-exec.c - cgi-exec preprocess data from HTML forms and redirects
 *              the data to a specified CGI script responsible for
 *              processing the data from that particular document.
 *
 * Copyright (c) 2000, 2002, 2003 William F. Simmons.  All rights reserved.
 *
 *    *    *    *    *    *    *    *    *    *    *    *    *    *    *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to:
 *
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330
 * Boston, MA 02111-1307
 * USA
 *
 *    *    *    *    *    *    *    *    *    *    *    *    *    *    *
 *
 * The author of this program can be reached at the following address:
 *
 * William F. Simmons
 * Simtech
 * 6427 Windfern
 * Houston, Texas 77040
 * USA
 *
 * Telephone: 713-460-5921
 * EMail:     wfs@simtech-soft.com
 * Web Site:  www.simtech-soft.com
 *
 *    *    *    *    *    *    *    *    *    *    *    *    *    *    *
 */

/*
 * cgi-exec.c  -  acquire all cgi input, regardless of method, decode input,
 *                and place input into environmental variables corresponding
 *                to the fieldnames designated in the html form; then execute
 *                the script cited in the cgi-exec.db database.
 *                overwriting existing environmental variables is illegal.
 *
 * Copyright (c) 2000, 2002, 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/2/2000
 */


char progname[] = "cgi-exec 1.2.4";
char copyright[] = "Copyright (c) 2000, 2002, 2003 William F. Simmons.  All rights reserved.";
char gpl[] = "This software comes with ABSOLUTELY NO WARRANTY;\n\
  for details see the GNU Public License, version 2.\n\
  This is free software, and you are welcome to redistribute it\n\
  under the terms of the license named above.\n";

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#define CGI_SEARCHPATH   "../../cgi-exec:../cgi-exec:"
#define STDIN               0
#define MAXLINE           128
#define MAXFLDS            32
#define PATHLEN            96
#define MSGLEN            128
#define DBFLDS              3


/*
 * cgi_fatal.c - send error message and die.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/21/94, 12/4/95, 12/7/95, 5/10/2000, 8/3/2000
 */


#ifndef CGI_FATALVAL
#define CGI_FATALVAL    0
#endif


cgi_fatal(message, header_type)

char *message;
int header_type;

{
  extern char progname[];
  extern char copyright[];

  /* print header */
  if(header_type)
    {
      if(header_type == 1) printf("content-type: text/plain\n\n");
      if(header_type == 2) printf("content-type: text/html\n\n<html>\n");
    }

  /* print message */
  if(header_type == 2)
    printf("<p>\n%s. %s<br>\n%s<br>\n<p>\n", progname, copyright, message);
  else
    printf("%s. %s\n%s\n", progname, copyright, message);

  /* print footer */
  if(header_type)
    {
      if(header_type == 1) printf("\n");
      if(header_type == 2) printf("</html>\n");
    }

  fflush(stdout);

  exit(CGI_FATALVAL);
}



/*
 * upcase.c - convert string to upper case.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/10/95, 12/4/95, 8/2/98
 */


upcase(string)

char *string;

{
  register char *str;

  str = string;
  while(*str)
    {
      if(*str >= 'a' && *str <= 'z')
        *str -= 32;
      str++;
    }
}



/*
 * zapflds.c - overwrite characacter pointed at in fieldends with zapchar
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/2/2000, 3/13/2003
 */

void zapflds(fieldends, zapchar)

char **fieldends;
char zapchar;

{
  while(*fieldends)
    {
      **fieldends =  zapchar;
      fieldends++;
    }
}



/*
 * decodecgifld.c - convert cgi whitespace encodement to text counterpart
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 8/2/1998, 11/14/2000
 */

int decodecgifld(enc)

char *enc;                                        /* cgi-encoded string */

{
  int nchars;

  nchars = 0;
  while(*enc)
    {
      nchars++;
      if(*enc == '&') *enc = '\n';
      if(*enc == '+') *enc = ' ';
      enc++;
    }

  return(nchars);
}



/*
 * decodecgihex.c - convert cgi hex to plain-text equivalent during copy.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/31/1998, 3/24/2000, 5/10/2000, 11/14/2000
 */

/* this assumes success. not good. */

int decodecgihex(enc,dec)

char *enc;              /* cgi-encoded source string */
char *dec;              /* decoded plain-text destination string */

{
  char ch;
  int nchars;

  nchars = 0;

  while(*enc)
    {
      if(*enc != '%')
        {
          *dec = *enc;
          nchars++;
        }
      else
        {
          enc++;

          if(*enc >= '0' && *enc <= '9')
            ch = (*enc - '0') * 16;
          if(*enc >= 'A' && *enc <= 'Z')
            *enc +=32;
          if(*enc >= 'a' && *enc <= 'z')
            ch = ((*enc - 'a') + 10) * 16;

          enc++;

          if(*enc >= '0' && *enc <= '9')
            ch += (*enc - '0');
          if(*enc >= 'A' && *enc <= 'Z')
            *enc +=32;
          if(*enc >= 'a' && *enc <= 'z')
            ch += ((*enc - 'a') + 10);

          *dec = ch;
          nchars++;
        }

      enc++;
      dec++;
    }

  *dec = (char) 0;

  return(nchars);
}



/*
 * cgiparseinput.c - parse cgi input; map out starts and ends of names
 *                   and values.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/2/2000, 3/20/2000, 11/14/2000
 */

int cgiparseinput(encode_in,namestarts,nameends,valstarts,valends)

char *encode_in;      /* cgi-encoded input buffer */
char **namestarts;    /* beginnings of var names */
char **nameends;      /* ends of var names */
char **valstarts;     /* beginnings of var values */
char **valends;       /* ends of var values */

{
  register char *enc;
  register char **nmstart;
  register char **nmend;
  register char **valstart;
  register char **valend;
  register int nfields;

  enc = encode_in;
  nmstart = namestarts;
  nmend = nameends;
  valstart = valstarts;
  valend = valends;
  nfields = 0;

  *nmstart++ = enc;
  while(*enc)
    {
      if(*enc == '&')
        {
          *valend++ = enc;
          *nmstart++ = enc + sizeof(char);
          nfields++;
        }
      if(*enc == '=')
        {
          *nmend++ = enc;
          *valstart++ = enc + sizeof(char);
        }
      enc++;
    }
  if(enc > encode_in)
    {
      *valend++ = enc;
      nfields++;
    }

  *nmstart = (char *) 0;
  *nmend = (char *) 0;
  *valstart = (char *) 0;
  *valend = (char *) 0;

  return(nfields);
}



/*
 * cgidecodeinput.c - decode input and copy to output buffer;
 *                    map beginnings and ends of fieldnames and
 *                    values in output buffer.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/24/2000, 5//2000, 11/14/2000, 12/2/2000, 3/8/2003
 */


#define DATA_GAP     2
#define ENUM_LEN     2
#define SAFETY_GAP   ENUM_LEN + DATA_GAP


int cgidecodeinput(inputbuf,outputbuf,namestarts_in,namestarts_out,valstarts_in,valstarts_out)

char *inputbuf;          /* cgi-encoded input source buffer */
char *outputbuf;         /* decoded plain-text output buffer */
char **namestarts_in;    /* beginnings of var names in input buffer */
char **namestarts_out;   /* beginnings of var names in output buffer */
char **valstarts_in;     /* beginnings of var values in input buffer */
char **valstarts_out;    /* beginnings of var values in output buffer */

{
  char *output;
  char *input;
  char **nmstart_in;
  char **nmstart_out;
  char **valstart_in;
  char **valstart_out;
  int len;
  int nwritten;

  output = outputbuf;
  input = inputbuf;
  nmstart_in = namestarts_in;
  valstart_in = valstarts_in;
  nmstart_out = namestarts_out;
  valstart_out = valstarts_out;
  nwritten = 0;

  while(*nmstart_in)
    {
      len = decodecgifld(*nmstart_in);
      *nmstart_out = output;
      len = decodecgihex(*nmstart_in, output);
      *(output + len) = 0;
      output += len;
      output += SAFETY_GAP;   /* safety gap related to enumerator length */
      nwritten += len;
      nmstart_in++;
      nmstart_out++;

      len = decodecgifld(*valstart_in);
      *valstart_out = output;
      len = decodecgihex(*valstart_in, output);
      *(output + len) = 0;
      output += len;
      output += DATA_GAP;                    /* this gap does not change */
      nwritten += len;
      valstart_in++;
      valstart_out++;
    }

  *nmstart_out = (char *) 0;
  *valstart_out = (char *) 0;

  return(nwritten);
}



/*
 * cgicountflds.c - examine fieldnames for uniqueness; determine
 *                  how many repetitions of fieldname exist.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 4/21/2000, 5/21/2000, 11/14/2000, 11/16/2002, 3/13/2003
 */

cgicountflds(namestarts_out, namecounts, sequence)

char **namestarts_out;    /* beginnings of var names in output buffer */
int *namecounts;          /* number of occurrences of varname */
int *sequence;            /* value of suffix id for current varname */

{
  char **nmstart_out;
  char **thisname;
  int *nmcount;
  int *seq;

  thisname = namestarts_out;
  nmcount = namecounts;
  seq = sequence;

  while(*thisname)
    {
      nmstart_out = namestarts_out;
      *nmcount = 0;
      *seq = 0;
      while(*nmstart_out)
        {
          if(strcmp(*thisname,*nmstart_out) == 0)
            {
              ++*nmcount;
              if(nmstart_out <= thisname)
                ++*seq;
            }
          nmstart_out++;
        }
      seq++;
      nmcount++;
      thisname++;
    }

  *nmcount = 0;
  *seq = 0;
}



/*
 * cgifldfmt.c - construct enumerator format string for fieldnames;
 *               generate enumerators, and append onto non-unique
 *               fieldnames.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 4/21/2000, 5/21/2000, 11/14/2000, 12/2/2000
 */

cgifldfmt(namestarts_out, namecounts, sequence)

char **namestarts_out;    /* beginnings of varnames in output buffer */
int *namecounts;          /* number of occurrences of varname */
int *sequence;            /* value of suffix id for current varname */

{
  char **nmstart_out;
  char *formatstring;
  char scratchstr[64];
  int *nmcount;
  int *seq;

  nmstart_out = namestarts_out;
  nmcount = namecounts;
  seq = sequence;

  while(*nmcount)
    {
      if(*nmcount > 1)
        {
          formatstring = "%s%02.2d";     /*see safety gap in cgidecodeinput()*/
          sprintf(scratchstr, formatstring, *nmstart_out, *seq);
          strcpy(*nmstart_out,scratchstr);
        }
      nmstart_out++;
      seq++;
      nmcount++;
    }
}



/*
 * cgienvpopulate.c - populate environment with fields in output buffer.
 *                    overwriting existing variables is not allowed.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 5/2/2000, 5/18/2000, 11/14/2000
 */

cgienvpopulate(namestarts_out,valstarts_out)

  char **namestarts_out;     /* name of environmental variable to set */
  char **valstarts_out;      /* value of environmental variable to set */

{
  char **nmstart_out;
  char **valstart_out;

  nmstart_out = namestarts_out;
  valstart_out = valstarts_out;

  while(*nmstart_out)
    {
      if(getenv(*nmstart_out) != NULL)
        {
          printf("envronmental variable [%s] already exists\n", *nmstart_out);
          exit(0);
        }
      else
        {
          if(setenv(*nmstart_out, *valstart_out, 0) == -1)
            {
              printf("not enough room in environment to set [%s]\n",
                      *nmstart_out);
              exit(0);
            }
        }
      nmstart_out++;
      valstart_out++;
    }
}



/*
 * readline.c - read a single line of text from a file
 *
 * thanks to dennis ritchie and brian kernighan. (slightly modified)
 * wfs, houston, 6/21/2000, 6/27/2000
 */


int readline(line, limit, fp)

char *line;
int limit;
FILE *fp;

{
  register char *curchar;
  register int ch;
  register int nread;

  nread = 0;
  curchar = line;
  limit--;

  while((nread < limit) && ((ch = getc(fp)) != EOF)  &&  (ch != '\n'))
    {
      *curchar++ = ch;
      nread++;
    }

  if(ch == '\n')
    {
      *curchar++ = ch;
      nread++;
    }
  *curchar = '\0';

  return(nread);
}



/*
 * cgigetreferer.c - find http_referer; parse string and return only
 *                   local document name.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 6/28/2000, 11/14/2000
 */

char *cgigetreferer()

{
  char *http_referer;
  char *httpreferer;
 
/*
 * that's right. "referrer" is misspelled.
 */
  if((http_referer = getenv("HTTP_REFERER")) == NULL)
    return(http_referer); /*NULL*/

/*
 * to finish parsing http_referer, test for "^http://" then skip to first
 * character past the first "/".  if no "http://" found, go to last letter
 * and come back to the first "/", try to stat $DOCUMENT_ROOT/http_referer.
 * if not found, go back to next "/" and try again until stat works or
 * beginning of string encountered. if beginning of string, fail.
 */

  httpreferer = http_referer;
  if((http_referer = (char *) strstr(http_referer, "http://")) == NULL)
    {
      return(http_referer); /*NULL: http_referer does not contain "http://"*/
    }
  else
    {
      http_referer = (char *) http_referer + 7;
      http_referer = (char *) strchr(http_referer, '/');
      http_referer++;
    }

  return(http_referer);
}



/*
 * getscriptdir.c - parse internal searchpath, build possible pathnames,
 *                  and test for existence of resulting directories.
 *                  return path of first-matched directory, or of pwd
 *                  if no directories are matched.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 2/27/2003, 2/28/2003, 3/3/2003
 */

char *getscriptdir(pwd, cgi_searchpath, cgi_scriptdir)

char *pwd;
char *cgi_searchpath;
char *cgi_scriptdir;

{
  char *starts[8];
  char *ends[8];
  char *strt;
  char *end;
  char scratchdir[PATHLEN];
  char *srchpath;
  char *scratch;
  int npaths;
  int pathno;

  /* parse searchpath */
  npaths = 0;
  srchpath = cgi_searchpath;
  starts[npaths] = srchpath;
  scratch = NULL;

  while((scratch = (char *) strchr(srchpath, (int) ':')) != NULL)
    {
      ends[npaths] = scratch;
      npaths++;
      starts[npaths] = (scratch + sizeof(char));
      srchpath = starts[npaths];
    }

  starts[npaths] = (char *) 0;
  ends[npaths] = (char *) 0;

  pathno = 0;
  while(pathno < npaths)
    *ends[pathno++] = '\0';

  /* examine each searchpath entry */
  pathno = 0;
  while(pathno < npaths)
    {
      /* build absolute pathname from searchpath entries */
      if(*starts[pathno] == '/')
        {
          sprintf(cgi_scriptdir, "%s", starts[pathno]);
        }
      else
        {
          sprintf(scratchdir, "%s", pwd);

          strt = starts[pathno];
          while((scratch = (char *) strstr(strt, "../")) != NULL)
            {
              end = (char *) rindex(scratchdir, (int) '/');
              *end = '\0';
              strt += 3;
            }
          sprintf(cgi_scriptdir, "%s/%s", scratchdir, strt);

          if((scratch = (char *) strstr(strt, "./")) != NULL)
            {
              strt += 2;
            }

          if(*strt)
            sprintf(cgi_scriptdir, "%s/%s", scratchdir, strt);
          else
            sprintf(cgi_scriptdir, "%s", scratchdir);
        }

      /* test for existence of directory; break if found */
      if((access(cgi_scriptdir, F_OK)) == 0)
        return(cgi_scriptdir);

    pathno++;
  }


  return(pwd);
}


/*
 * getpwd.c - get the name of the current working directory.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 6/30/2000, 2/28/2003
 */

int getpwd(pwd)

char *pwd;

{
  int len;

  len = PATHLEN;

  if(getcwd(pwd, len) == NULL)
    return(0);

  len = strlen(pwd);
  if(pwd[len - 1] == '/')
    pwd[len] = '\0';

  return(1);
}



/*
 * getcgiexec.c - get the name of the currently executing program;
 *                return only the basename of that program.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 6/30/2000
 */

/*
 * SCRIPT_FILENAME appears not to be the way to find the path of
 * cgi-exec, since SCRIPT_FILENAME is not always present, and
 * sometimes is not the name of the cgi program.  
 *
 * the path of cgi-exec appears to be consistently available as a
 * combination of PWD and a fragment of SCRIPT_NAME or REQUEST_URL.
 * PATH_INFO and PATH_TRANSLATED seem to be candidates also, as do
 * REDIRECT_SCRIPT_URI and REDIRECT_SCRIPT_URL.  but the last four
 * seem not to be consistently available.
 *
 *  environmental variable PWD is not available to getenv until
 *  /bin/sh, et al are spawned.  i had to use the getcwd system call
 *  instead.
 */

char *getcgiexec()

{
  char *script_name;
  char *request_url;
  char *cgi_exec;

  script_name = NULL;
  request_url = NULL;
  cgi_exec = NULL;

  if((script_name = getenv("SCRIPT_NAME")) == NULL)
    {
      if((request_url = getenv("REQUEST_URL")) == NULL)
        return(NULL);
      else
        {
          request_url = (char *) rindex(request_url, '/');
          request_url++;
          cgi_exec = request_url;
        }
    }
  else
    {
      script_name = (char *) rindex(script_name, '/');
      script_name++;
      cgi_exec = script_name;
    }

  return(cgi_exec);
}



/*
 * getcgirecord.c - search the cgi-exec.db database for the record containing
 *                  the string pointed to by http_referer in the first field.
 *                  return pointer to that string, if record found; null
 *                  if no record found.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/2/2000, 10/27/2000, 1/5/2003, 2/6/2003
 */

char *getcgirecord(cgi_exec_db, http_referer, line)

char *cgi_exec_db;    /* full pathname of db file (cgi-exec.db) */
char *http_referer;   /* value of HTTP_REFERRER. 1st field in db record */
char *line;           /* scratch buffer to examine db records */

{
  int linelen;
  FILE *fp;
  char *cgi_record;
  char *scratchptr;

  if(http_referer  == NULL)
    return(NULL);  /* no search string */

  if((fp = fopen(cgi_exec_db, "r")) == NULL)
    return(NULL);  /* cannot open cgi-exec.db */

  cgi_record = NULL;
  while((linelen = readline(line, MAXLINE, fp)) > 0)
    {
      if((scratchptr = (char *) strstr(line, http_referer)) == line)
        {
          cgi_record = line;
          break;
        }
    }

  fclose(fp);

  return(cgi_record);
}



/*
 * parsecgirecord.c - parse record from cgi-exec.db into component fields.
 *                    return number of fields found.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 2/21/2003
 */

int parsecgirecord(line, starts, ends)

char line[];
char *starts[];
char *ends[];

{
  int nflds;
  char *scratch;

  nflds = 0;
  scratch = NULL;

  starts[nflds] = line;

  while((scratch = (char *) strchr(line, (int) '\t')) != NULL)
    {
      ends[nflds] = scratch;
      nflds++;
      starts[nflds] = (scratch + sizeof(char));
      line = starts[nflds];
    }

  if((scratch = (char *) strchr(line, (int) '\n')) == NULL)
    return(-1);    /* no newline in record: bad */
  else
    ends[nflds++] = scratch;

  starts[nflds] = (char *) 0;
  ends[nflds] = (char *) 0;

  return(nflds);
}


/*
 * getcgiscriptname.c - check if cgi_script is relative or full path.
 *                      if full path, filename is absolute.
 *                      if relative prepend pwd onto cgi_script.
 *                      in either case, return full pathname of script
 *                      to be spawned.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/2/2000
 */

char *getcgiscriptname(cgi_dir, cgi_script, scratchspace)

char *cgi_dir;
char *cgi_script;
char *scratchspace;

{
  char *cgi_scriptname;
  char *scratchptr;

  scratchptr = (char *) strchr(cgi_script, (int) '/');
  if(! scratchptr)
    {
      /* relative pathname. */
      sprintf(scratchspace, "%s/%s", cgi_dir, cgi_script);
      cgi_scriptname = scratchspace;
    }
  else
    {
     /* absolute pathname. */
      if(scratchptr == cgi_script)
        cgi_scriptname = cgi_script;
      /* bad pathname. who knows? */
      else
        return(NULL);
    }

  return(cgi_scriptname);
}



/*
 * cgirequestmethod.c - return the request_method of current transaction.
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/2/2000, 12/2/2000
 */

/* this is a stupid function. wfs. */
/* consider doing directly in main() or inline macro */

char *cgirequestmethod()

{
  char *request_method;

  if((request_method = getenv("REQUEST_METHOD")) == NULL)
    return(request_method);/*NULL*/

  return(request_method);
}



/*
 * printhead.c - print appropriate header type
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 8/3/2000
 */

printhead(header_type)

int header_type;

{
  /*
   * header_type == 0: no header
   * header_type == 1: text/plain
   * header_type == 0: text/html
   */

  if(header_type)
    {
      if(header_type == 1)
        printf("content-type: text/plain\n\n");
      if(header_type == 2)
        printf("content-type: text/html\n\n<html>\n");  
    }
  fflush(stdout);
}



/*
 * printfoot.c - print appropriate footer type
 *
 * Copyright (c) 2000 William F. Simmons.  All rights reserved.
 * wfs, houston, 8/3/2000
 */

printfoot(header_type)

int header_type;

{
  /*
   * header_type == 0: no header  / no footer
   * header_type == 1: text/plain / no footer
   * header_type == 0: text/html  / html footer
   */

  if(header_type)
    {
      if(header_type == 1) printf("\n");
      if(header_type == 2) printf("</html>\n");  
    }
  fflush(stdout);
}



/*
 * bufalloc.c - acquire input in increments of blksize until input exhausted,
 *              or until maxbuf is exceeded.  return pointer to input buffer
 *              if successful; null pointer if not.  write number of bytes
 *              read into parameter "count": 0 if no bytes read, in which
 *              case input pointer is null, since there is no input; otherwise
 *              -1 if a read error occurred; -2 if allocating memory was the
 *              problem, or -3 if exceeding maxbuf was the problem.
 *              both alloc and maxbuf errors return null pointers.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 20030306, 20030307
 */


#ifndef _UNISTD_H
#include <unistd.h>
#endif

#define BLKSIZE       256
#define MAXBUF        BLKSIZE * 8
#define BUF_READ      -1
#define BUF_ALLOC     -2
#define BUF_EXCEED    -3


char *bufalloc(count)

int *count;

{
  char *buffer;
  char *buf;
  int nread;
  int totalread;
  int ntries;

  nread = 0;
  totalread = 0;
  ntries = 0;


  if((buffer = malloc(BLKSIZE)) == NULL)
    {
      nread = BUF_ALLOC;
      goto check_input;
    }

  ntries++;
  buf = buffer;


  while((nread = read(STDIN, buf, BLKSIZE)) == BLKSIZE)
    {
      ntries++;

      if((BLKSIZE*ntries) > MAXBUF)
        {
          nread = BUF_EXCEED;
          break;
        }

      if((buffer = realloc(buffer, (BLKSIZE*ntries))) == NULL)
        {
          nread = BUF_ALLOC;
          break;
        }

      totalread += nread;
      buf = (buffer + totalread);
    }


  check_input:

  if(nread < 0)
    {
      totalread = nread;
    }
  else
    {
      totalread += nread;
      *(buffer + totalread) = 0;
    }

  if(totalread <= 0 && buffer != NULL)
    {
      free(buffer);
      buffer = NULL;
    }


  *count = totalread;
  return(buffer);
}



/*
 * cgigetinput.c - check request_method; if "get" then point to query_string;
 *                 if "post" then allocate memory and read input into buffer.
 *                 return pointer to the beginning of the input data.
 *
 * Copyright (c) 2000, 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 7/2/2000, 12/2/2000, 3/7/2003
 */

char *cgigetinput(request_method, inputlen)

char *request_method;
int *inputlen;

{
  char *query_string;
  char *input;
  int length;

  upcase(request_method);

  if(strcmp(request_method, "GET") == 0)
    {
      if((query_string = getenv("QUERY_STRING")) == NULL)
        cgi_fatal("cannot access QUERY_STRING");
      input = query_string;
      length = strlen(query_string);
    }

  if(strcmp(request_method, "POST") == 0)
    {
      if((input = bufalloc(&length)) == NULL)
        {
          if(length < 0)
            {
              if(length == BUF_READ)
                cgi_fatal("bufalloc(): read error");
              if(length == BUF_ALLOC)
                cgi_fatal("bufalloc(): memory allocation error");
              if(length == BUF_EXCEED)
                cgi_fatal("bufalloc(): memory allocation exceeds permissible size");
            }
        }
    }

  *inputlen = length;

  return(input);
}



/*
 * cgioutputbuf.c - hide the ugliness of computing the length of the
 *                  output buffer.  return pointer to buffer;
 *                  store length of buffer in outputlen.
 *                  in practice the output data is normally shorter
 *                  than the input data; but it could happen that
 *                  all input data is plain text, in which case the enumerator
 *                  safety gaps become an issue.  this function allocates
 *                  its buffer using the worst-case scenario in which no
 *                  cgidecoding was necessary.
 *
 * cf. cgidecodeinput(), cgifldfmt()
 *
 * Copyright (c) 2003, William F. Simmons.  All rights reserved.
 * wfs, houston, 20030308
 */


char *cgioutputbuf(inputlen, nrecs, outputlen)

int inputlen;
int nrecs;
int *outputlen;

{
  char *output;
  int length;

  /* already conservative, (SAFETY_GAP * 2) was included just in case */
  length = (inputlen + (nrecs * SAFETY_GAP) + (nrecs * DATA_GAP) + (SAFETY_GAP * 2));

  if((output = malloc(length)) == NULL)
    length = 0;

  *outputlen = length;
  return(output);
}



/*
 * cgigetflds.c - count the number of input fields in cgi input buffer;
 *                return number of fields found, or 0 if none found.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 20030312
 */

int cgigetflds(input)

char *input;

{
  char *curchar;
  int nfields;

  nfields = 0;
  curchar = input;

  while(*curchar)
    {
      if(*curchar == '&')
        nfields++;

      curchar++;
    }

  if(curchar > input)
    nfields++;

  return(nfields);
}



/*
 * cgiparsebuf. c - allocate parsing buffer; compute size according
 *                  to number of fields found in input.  return
 *                  pointer to allocated buffer.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 20030312
 */


char **cgiparsebuf(nfields)

int nfields;

{
  char **buf;
  int size;

  /* better safe than sorry; hence +2 instead of +1 */
  size = ( (nfields * sizeof(char *)) + (2 * sizeof(char *)) );

#ifdef CGI_DEBUG
printf("cgiparsebuf(): size = [%d]\n", size);
fflush(stdout);
#endif

  if((buf = malloc(size)) == NULL)
    cgi_fatal("cgiparsebuf(): could not allocate parsebuf");

  return(buf);
}



/*
 * cgiintbuf. c - allocate integer buffer; compute size according
 *                to number of fields found in input.  return
 *                pointer to allocated buffer.
 *
 * Copyright (c) 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 20030313
 */


int *cgiintbuf(nfields)

int nfields;

{
  int *buf;
  int size;

  /* better safe than sorry; hence +2 instead of +1 */
  size = ( (nfields * sizeof(int *)) + (2 * sizeof(int *)) );

#ifdef CGI_DEBUG
printf("cgiintbuf(): size = [%d]\n", size);
fflush(stdout);
#endif

  if((buf = malloc(size)) == NULL)
    cgi_fatal("cgiintbuf(): could not allocate integer buffer");

  return(buf);
}



/*
 * cgi-exec.c
 * Copyright (c) 2000, 2002, 2003 William F. Simmons.  All rights reserved.
 * wfs, houston, 3/2/2000, 3/24/2000,
 *               4/26/2000,
 *               5/2/2000, 5/9/2000, 5/18/2000, 5/21/2000,
 *               6/20/2000, 6/21/2000, 6/26/2000, 6/27/2000,
 *               6/28/2000, 6/29/2000, 6/30/2000,
 *               7/2/2000,
 *               8/3/2000,
 *               9/20/2000, 9/21/2000, 9/22/2000,
 *               10/27/2000,
 *               11/30/2000,
 *               12/1/2000, 12/2/2000,
 *               11/16/2002,
 *               1/5/2003,
 *               1/28/2003,
 *               2/6/2003,
 *               2/21/2003,
 *               2/27/2003, 2/28/2003,
 *               3/7/2003, 3/8/2003,
 *               3/12/2003, 3/13/2003, 3/14/2003, 3/15/2003
 */

main(argc, argv, envp)

int argc;
char *argv[];
char *envp[];

{
  char *request_method;
  char *http_referer;
  char *ref_doc;
  char cgi_searchpath[sizeof(CGI_SEARCHPATH) + 1];
  char cgi_dir[PATHLEN];
  char script_dir[PATHLEN];
  char *cgi_exec;
  char cgi_exec_db[PATHLEN];
  char *cgi_record;
  char *scrptdir;
  char *cgi_script;
  char *cgi_scriptname;
  char *header;
 
  char *input;
  int  inputlen;

  char **namestarts_in;
  char **nameends_in;
  char **valstarts_in;
  char **valends_in;

  char *output;
  int  outputlen;

  char **namestarts_out;
  char **valstarts_out;

  int *namecounts;
  int *sequence;

  int header_type;

  char line[MAXLINE];
  int nwritten;
  int nrecs;
  int nflds;

  char *fieldstarts[DBFLDS + 1];
  char *fieldends[DBFLDS + 1];

  char message[MSGLEN];
  char scratchspace[PATHLEN];

  char zapchar;
  zapchar = '\0';

  strcpy(cgi_searchpath, CGI_SEARCHPATH);

/***** process args *****/

#ifdef CGI_DEBUG
printf("content-type: text/plain\n\n");
printf("%s\n", progname);
printf("%s\n", copyright);
printf("%s\n\n", gpl);
#endif

  header_type = 1;  /* plain text header by default */

  if(argc != 1)
    {
      sprintf(message, "cgi-exec does not accept any command-line parameters");
      cgi_fatal(message, header_type);
    }


/***** get referrer *****/

  http_referer = NULL;
  http_referer = cgigetreferer();

#ifdef CGI_DEBUG
printf("http_referer: [%s]\n", http_referer);
fflush(stdout);
#endif


/***** get cgi-exec *****/

  if(! getpwd(cgi_dir))
    cgi_fatal("cannot get current working directory (PWD)", header_type);

#ifdef CGI_DEBUG
printf("cgi_dir: [%s]\n", cgi_dir);
fflush(stdout);
#endif

  if((cgi_exec = getcgiexec()) == NULL)
    cgi_fatal("cannot get cgi_exec", header_type);

#ifdef CGI_DEBUG
printf("cgi_exec: [%s]\n", cgi_exec);
fflush(stdout);
#endif


/***** get cgi-exec.db *****/

  scrptdir =  getscriptdir(cgi_dir, cgi_searchpath, script_dir);

#ifdef CGI_DEBUG
printf("scriptdir: [%s]\n", scrptdir);
fflush(stdout);
#endif

  sprintf(cgi_exec_db, "%s/%s.db", scrptdir, cgi_exec);

#ifdef CGI_DEBUG
printf("cgi-exec.db: [%s]\n", cgi_exec_db);
fflush(stdout);
#endif

/*check existence of file. fail if not found*/
  if((access(cgi_exec_db, F_OK)) < 0)
    cgi_fatal("cannot open cgi-exec.db", header_type);

/*check permissions. fail if writable*/
  if((access(cgi_exec_db, W_OK)) == 0)
    cgi_fatal("cgi-exec.db is writable; failing", header_type);

#ifdef CGI_DEBUG
printf("cgi-exec.db ok: [%s]\n", cgi_exec_db);
fflush(stdout);
#endif


/***** allocate input data buffers, get html form input *****/

/*
 * wfs, 20030207: this needs more thought.
 * wfs, 20030308: this claim is not true!
 *                request_method is defined even for an anchor tag!
 *                the issue is not whether or not there is a
 *                request_method defined, but whether there is
 *                any input.
 *                probably the best thing to do is to get the
 *                request_method, and apply no tests or jumps.
 * original (errant) notion:
 * if there is no request_method, then there is no <form> in the 
 * referring document, and a backup searchkey into cgi-exec.db via
 * ref_doc cannot be passed along.  so the search will have to be made
 * on the strength of http_referer, an unreliable switching mechanism.
 * this means that a page without a form (ie, a normal link) can
 * only support the execution of a single cgi-script with cgi-exec.
 * the jump below may need to be changed.
 *
 * if there is no request_method, there is no input: just execute script.
 */

 if((request_method = cgirequestmethod()) == NULL)
   goto search_table;

#ifdef CGI_DEBUG
printf("request_method: [%s]\n", request_method);
fflush(stdout);
#endif

  input = cgigetinput(request_method, &inputlen);

#ifdef CGI_DEBUG
printf("cgigetinput(): ok\n");
printf("cgigetinput(): bytes read: [%d]\n", inputlen);
printf("cgigetinput(): buffer contents:\n [%s]\n", input);
fflush(stdout);
#endif

/*
 * allow input to be null:
 * page has no <form> statement, or
 * <form> has no data fields
 */

  if(inputlen == 0)
    {

#ifdef CGI_DEBUG
printf("cgigetinput(): no input; skipping to search_table\n");
fflush(stdout);
#endif

      goto search_table;

    }
  else
    {

#ifdef CGI_DEBUG
printf("cgigetinput(): input acquired; proceed to output buffer allocation and parsing functions\n");
fflush(stdout);
#endif

      ; /* nothing to do */

    }


/***** allocate input/output parsing buffers *****/

  nrecs = cgigetflds(input);

#ifdef CGI_DEBUG
printf("cgigetflds(): ok. nfields = [%d]\n", nrecs);
fflush(stdout);
#endif

  if(nrecs == 0)
    goto search_table;

  if(nrecs > MAXFLDS)
    cgi_fatal("cgigetflds(): maximum number of fields from form exceeded");

  /* allocate parsing buffers */
  namestarts_in  = cgiparsebuf(nrecs);
  nameends_in    = cgiparsebuf(nrecs);
  valstarts_in   = cgiparsebuf(nrecs);
  valends_in     = cgiparsebuf(nrecs);
  namestarts_out = cgiparsebuf(nrecs);
  valstarts_out  = cgiparsebuf(nrecs);

  namecounts     = cgiintbuf(nrecs);
  sequence       = cgiintbuf(nrecs);


/***** parse form input *****/

  nrecs = cgiparseinput(
                        input,
                        namestarts_in, nameends_in,
                        valstarts_in, valends_in
                       );

#ifdef CGI_DEBUG
printf("cgiparseinput(): ok. nfields = [%d]\n", nrecs);
fflush(stdout);
#endif

  zapflds(nameends_in, zapchar);
  zapflds(valends_in, zapchar);


  if((output = cgioutputbuf(inputlen, nrecs, &outputlen)) == NULL)
    cgi_fatal("could not allocate output buffer", header_type);


#ifdef CGI_DEBUG
printf("cgioutputbuf(): ok. buffer size = [%d]\n", outputlen);
fflush(stdout);
#endif

  nwritten = cgidecodeinput(
                            input, output,
                            namestarts_in, namestarts_out,
                            valstarts_in, valstarts_out
                           );


#ifdef CGI_DEBUG
printf("cgidecodeinput(): ok\n");
printf("cgidecodeinput(): bytes written into output buffer = [%d]\n", nwritten);
fflush(stdout);
#endif

  cgicountflds(namestarts_out, namecounts, sequence);

#ifdef CGI_DEBUG
printf("cgicountflds(): ok\n");
fflush(stdout);
#endif

  cgifldfmt(namestarts_out, namecounts, sequence);

#ifdef CGI_DEBUG
printf("cgifldfmt(): ok\n");
fflush(stdout);
#endif

  cgienvpopulate(namestarts_out,valstarts_out);

#ifdef CGI_DEBUG
printf("cgienvpopulate(): ok\n");
fflush(stdout);
#endif


/***** get cgi_record *****/

search_table:

  if((cgi_record = getcgirecord(cgi_exec_db, http_referer, line)) == NULL)
    {

#ifdef CGI_DEBUG
printf("getcgirecord() failed to find record in cgi-exec.db using http_referer = [%s]\n", http_referer);
fflush(stdout);
#endif

      /* 
       * REF_DOC should be put into all html forms as a hidden field.
       * It should be the name of the html document relative to DOCUMENT_ROOT.
       * This is a fail-safe device that will only be used if http_referer
       * fails to yield a script.  So, if someone is masking out their 
       * http_referer, a script can still be obtained using REF_DOC.
       * It also allows the construction of deep dynamic web applications,
       * since http_referer will always be:
       *   "http://host.domain.com/cgi-bin/cgi-exec"
       * after the first dynamic page.
       */

      if((ref_doc = getenv("REF_DOC")) == NULL)
        cgi_fatal("cannot find ref_doc", header_type);

#ifdef CGI_DEBUG
printf("ref_doc: [%s]\n", ref_doc);
fflush(stdout);
#endif

      if((cgi_record = getcgirecord(cgi_exec_db, ref_doc, line)) == NULL)
        cgi_fatal("cannot find record in cgi-exec.db", header_type);

#ifdef CGI_DEBUG
printf("getcgiscript() found record using ref_doc = [%s]\n", ref_doc);
fflush(stdout);
#endif

    }
  else
    {

#ifdef CGI_DEBUG
printf("getcgirecord() found record using http_referer = [%s]\n", http_referer);
fflush(stdout);
#endif

      ; /*do nothing*/
    }

#ifdef CGI_DEBUG
printf("matching record in cgi-exec.db:\n%s", cgi_record);
fflush(stdout);
#endif


/***** parse cgi_record *****/

if((nflds = parsecgirecord(line, fieldstarts, fieldends)) != DBFLDS)
  {
    /* wrong number of fields in cgi-exec.db record */
    sprintf(message, "malformed record in cgi-exec.db. nflds = [%d]", nflds);
    cgi_fatal(message, header_type);
  }

zapflds(fieldends, zapchar);

cgi_script = fieldstarts[1];
header = fieldstarts[2];

header_type = 1;  /* default to text/plain */

if(strcmp(header, "none") ==  0)
  header_type = 0;
if(strcmp(header, "text") ==  0)
  header_type = 1;
if(strcmp(header, "html") ==  0)
  header_type = 2;

#ifdef CGI_DEBUG
printf("header type: [%s]\n", header);
fflush(stdout);
#endif


/***** get cgi_script *****/

#ifdef CGI_DEBUG
printf("cgi_script: [%s]\n", cgi_script);
fflush(stdout);
#endif

  if((cgi_scriptname=getcgiscriptname(scrptdir,cgi_script,scratchspace))==NULL)
    cgi_fatal("cannot get cgi_scriptname", header_type);

#ifdef CGI_DEBUG
printf("getcgiscriptname(): ok\n");
printf("cgi_scriptname: [%s]\n", cgi_scriptname);
fflush(stdout);
#endif

/* does cgi-script exist */
  if((access(cgi_scriptname, F_OK)) < 0)
    cgi_fatal("cgi-script does not exist", header_type);

#ifdef CGI_DEBUG
printf("cgi_scriptname [%s]:  file exists\n", cgi_scriptname);
fflush(stdout);
#endif

/* is cgi-script executable */
  if((access(cgi_scriptname, X_OK)) < 0)
    cgi_fatal("cannot execute cgi-script", header_type);

#ifdef CGI_DEBUG
printf("cgi_scriptname [%s]:  file executable\n", cgi_scriptname);
fflush(stdout);
#endif


/*** spawn script ***/

#ifdef CGI_DEBUG
printf("now executing cgi_scriptname [%s]:\n", cgi_scriptname);
printf("==================================\n");
fflush(stdout);
#endif

  printhead(header_type);
  system(cgi_scriptname);
  printfoot(header_type);

#ifdef CGI_DEBUG
printf("==================================\n");
printf("cgi_scriptname [%s] exiting execution\n", cgi_scriptname);
fflush(stdout);
#endif

#ifdef CGI_DEBUG
printf("%s: terminating successfully\n", cgi_exec);
fflush(stdout);
#endif


  exit(0);
}


